UserParameter=user.discovery.docker-volume,sudo -u {{ hive_safe_admin }} sh -c "export DOCKER_HOST={{ groups['hives'] | intersect(groups[hive_stage]) | first }}:2376;export DOCKER_TLS_VERIFY=1;{{ hive_home_dir }}/docker/bin/python {{ hive_home_dir }}/docker-volume-discover.py"
UserParameter=docker-volume[*],sudo -u {{ hive_safe_admin }} sh -c "dexec -n $1 df $2 | awk 'NR==2{ sub(/.$/,\"\",\$$5);print \$$5;}'"
